import os
import sys
import csv
import numpy as np
import pandas as pd


def _deconstruct_git_move(fname):
    if "=>" in fname:
        if "{" in fname and "}" in fname:
            start, end = fname.split(" => ")
            new_piece, tail = end.split("}")
            head, old_piece = start.split("{")
            old = head + old_piece + tail
            new = head + new_piece + tail

            if "//" in old:
                old = old.replace("//", "/")
            if "//" in new:
                new = new.replace("//", "/")
        else:
            old, new = fname.split(" => ")

        return old, new
    else:
        raise Exception("Does not seem to be a file that was moved.")


mapping_seen = {}


def _saw_it_or_ancestor(new_name):
    if new_name in mapping_seen.keys():
        return True
    for line in mapping_seen.values():
        if new_name in line:
            return True
    return False


def _find_previous(data, el):
    # Finds the previous occurrence of el in data
    for idx, (new_name, old_name) in enumerate(data):
        if el == new_name:
            if idx + 1 < len(data):
                shorter_data = data[idx + 1 :]
            else:
                shorter_data = []
            return shorter_data, old_name
    return [], None


def _get_current_for(new_name):
    if new_name in mapping_seen.keys():
        return new_name
    for line in mapping_seen.values():
        if new_name in line:
            return line[0]
    # Should never happen
    return new_name


def repair(log_csv_name):
    df = pd.read_csv(log_csv_name, parse_dates=["date"], na_values=["-"])
    df["current"] = df.fname

    # Commits can be empty, so filter out the commits that do not affect files:
    df_files = df[pd.notna(df.fname)]

    # Create separate columns with old an new names for git mv actions
    df_moves = df_files[df_files.fname.str.contains(" => ")]

    olds_news_currents = []
    for fname in df.fname.values:
        if type(fname) is float:
            # Then it was an empty commit with the float signaling a nan value
            olds_news_currents.append((np.nan, np.nan, np.nan))
        elif "=>" in fname:
            old, new = _deconstruct_git_move(fname)
            olds_news_currents.append((old, new, np.nan))
        else:
            olds_news_currents.append((np.nan, np.nan, np.nan))

    olds, news, currents = list(zip(*list(olds_news_currents)))
    df["old"] = olds
    df["new"] = news
    df["current"] = currents

    # Find only the non-merge commits. They are those that have an `old` and
    # `new` field set but an empty `current` field.
    query = (pd.notna(df.old)) & (pd.notna(df.new)) & (pd.isna(df.current))
    nan_new = list(df[query].new.values)
    nan_old = list(df[query].old.values)

    new_to_old_zip = list(zip(nan_new, nan_old))

    for idx, new_name in enumerate(nan_new):
        if _saw_it_or_ancestor(new_name):
            # This is expensive ...
            continue
        line = []

        start_el = new_name
        shorter = new_to_old_zip[idx:]
        next_el = start_el
        while shorter and next_el:
            line.append(next_el)
            shorter, next_el = _find_previous(shorter, next_el)

        mapping_seen[line[0]] = line

    # And finally build a list that we can assign back to the DataFrame
    final_currents = []
    for _, row in df.iterrows():
        if pd.notna(row.old) and pd.notna(row.new) and pd.isna(row.current):
            final_currents.append(_get_current_for(row.new))
        elif pd.isna(row.old) and pd.isna(row.new) and pd.isna(row.fname):
            # Merging commits without any files attached
            final_currents.append(np.nan)
        elif (
            pd.isna(row.old)
            and pd.isna(row.new)
            and pd.notna(row.fname)
            and pd.isna(row.current)
        ):
            final_currents.append(_get_current_for(row.fname))
    df.current = final_currents

    out_path = log_csv_name.replace(".csv", "_repaired.csv")
    df.to_csv(out_path, index=False, quoting=csv.QUOTE_ALL)

    return out_path


if __name__ == "__main__":
    repair(sys.argv[1])



"""
This program computes the truck factor (also called bus factor) for a given 
Git repository.

It can be called like:

truckfactor <path_to_git_repository>

python -m truckfactor.run 

The truck factor specifies "...the number of people on your team that have to 
be hit by a truck (or quit) before the project is in serious trouble..."" 
L. Williams and R. Kessler, Pair Programming Illuminated

Note, do not call it from within a Git repository.

More on the truck factor:
  * https://legacy.python.org/search/hypermail/python-1994q2/1040.html
  * https://en.wikipedia.org/wiki/Bus_factor
"""
import os
import sys
import tempfile
import subprocess
from pathlib import Path
from shutil import which
from truckfactor.evo_log_to_csv import convert
from truckfactor.repair_git_move import repair
import pandas as pd


TMP = tempfile.gettempdir()


def write_git_log_to_file(path_to_repo):
    p = Path(path_to_repo)
    outfile = os.path.join(TMP, p.name + "_evo.log")

    cmd = f"""git -C {path_to_repo} log --pretty=format:'"%h","%an","%ad"' \
    --date=short \
    --numstat > \
    {outfile}"""

    subprocess.run(cmd, shell=True)

    return outfile


def preprocess_git_log_data(path_to_repo):
    evo_log = write_git_log_to_file(path_to_repo)
    evo_log_csv = convert(evo_log)
    evo_log_csv = repair(evo_log_csv)
    return evo_log_csv


def create_file_owner_data(df):
    new_rows = []

    for fname in set(df.current.values):
        view = df[df.current == fname]
        sum_series = view.groupby(["author"]).added.sum()
        view_df = sum_series.reset_index(name="sum_added")
        total_added = view_df.sum_added.sum()

        if total_added > 0:  # For binaries there are no lines counted
            view_df["owning_percent"] = view_df.sum_added / total_added
            owning_author = view_df.loc[view_df.owning_percent.idxmax()]
            new_rows.append(
                (
                    fname,
                    owning_author.author,
                    owning_author.sum_added,
                    total_added,
                    owning_author.owning_percent,
                )
            )
        # All binaries are silently skipped in this report...

    owner_df = pd.DataFrame(
        new_rows,
        columns=["artifact", "main_dev", "added", "total_added", "owner_rate"],
    )

    owner_freq_series = owner_df.groupby(["main_dev"]).artifact.count()
    owner_freq_df = owner_freq_series.reset_index(name="owns_no_artifacts")
    owner_freq_df.sort_values(by="owns_no_artifacts", inplace=True)

    return owner_df, owner_freq_df


def compute_truck_factor(df, freq_df):
    """Similar to G. Avelino et al.
    [*A novel approach for estimating Truck Factors*](https://ieeexplore.ieee.org/stamp)/stamp.jsp?arnumber=7503718)
    we remove low-contributing authors from the dataset as long as still more 
    than half of the files have an owner. The amount of remaining owners is the 
    bus-factor of that project.
    """
    no_artifacts = len(df.artifact)

    half_no_artifacts = no_artifacts // 2
    count = 0

    for owner, freq in freq_df.values:
        no_artifacts -= freq
        if no_artifacts < half_no_artifacts:
            break
        else:
            count += 1

    truckfactor = len(freq_df.main_dev) - count
    return truckfactor


def main(path_to_repo):
    evo_log_csv = preprocess_git_log_data(path_to_repo)
    complete_df = pd.read_csv(evo_log_csv)
    owner_df, owner_freq_df = create_file_owner_data(complete_df)
    truckfactor = compute_truck_factor(owner_df, owner_freq_df)
    return truckfactor


def git_is_available():
    """Check whether `git` is on PATH"""
    if which("git"):
        return True
    else:
        return False


def run():
    if not git_is_available():
        print(
            "Truckfactor requires `git` to be installed and accessible on path"
        )
        sys.exit(1)

    if len(sys.argv) <= 1:
        print(__doc__)
        sys.exit(1)

    path_to_repo = sys.argv[1]
    if not os.path.isdir(path_to_repo):
        print(__doc__)
        sys.exit(1)
    truckfactor = main(path_to_repo)

    print(f"The truck factor of {path_to_repo} is: {truckfactor}")


if __name__ == "__main__":
    run()



import os
import re
import csv
import sys
import tempfile
import subprocess


LINE_RE = r"(-|\d+)+\s+(-|\d+)+\s+(.*)"
RENAME_RE = r"\{(.*) => (.*)\}"
RENAME2_RE = r"(.*) => (.*)"


def parse_numstat_block(commit_line, block):
    if block:
        for line in block:
            # '-\t-\twww/static/screenshots/tree.png'
            m = re.match(LINE_RE, line)
            added, removed, file_name = m.groups()
            if added == "-":
                added = f'"{added}"'
            if removed == "-":
                removed = f'"{removed}"'

            csv_line = ",".join((commit_line, added, removed, f'"{file_name}"'))
            yield csv_line
    else:
        csv_line = ",".join((commit_line, "", "", ""))
        yield csv_line


def convert(report_file):
    with open(report_file) as fp:
        lines = fp.readlines()

    commit_blocks = []
    commit_block = []
    for idx, line in enumerate(lines):
        line = line.rstrip()
        if idx + 1 < len(lines):
            next_line = lines[idx + 1].rstrip()
        else:
            next_line = ""
        if line.startswith('"') and next_line.startswith('"'):
            # Next line is a commit too and they where no changes...
            commit_block.append(line)
            commit_blocks.append(commit_block[:])
            commit_block = []
        else:
            if line:
                commit_block.append(line)
            else:
                commit_blocks.append(commit_block[:])
                commit_block = []

    out_file = f"{report_file}.csv"
    out_path = os.path.join(tempfile.gettempdir(), out_file)
    with open(out_path, "w") as fp:
        fp.write("hash,author,date,added,removed,fname\n")
        for block in commit_blocks:
            commit_line = block[0]
            for csv_line in parse_numstat_block(commit_line, block[1:]):
                fp.write(csv_line + "\n")
    return out_path


if __name__ == "__main__":
    convert(sys.argv[1])
